"""A general class for holding solutions to pathway analyses."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Iterable, Optional

import numpy as np
import pandas as pd
from equilibrator_api import R, default_T, standard_concentration
from sbtab import SBtab

from .pathway import Pathway


class PathwayAnalysisSolution(object):
    """A general class for pathway analysis results."""

    def __init__(
        self,
        pathway: Pathway,
        score: float,
        ln_conc: np.ndarray,
        y: Optional[np.array] = None,
    ) -> None:
        """

        Parameters
        ----------
        thermo_model : ThermodynamicModel
        ln_conc : array
            log concentrations at MDF optimum
        """
        self.score = score
        self.ln_conc = ln_conc
        self.ln_conc_mu = pathway.ln_conc_mu
        self.ln_conc_sigma = pathway.ln_conc_sigma
        self.dg_confidence = pathway.dg_confidence
        self.ln_conc_confidence = pathway.ln_conc_confidence
        self.y = y

        standard_dg_primes = pathway.standard_dg_primes.squeeze()

        physiological_dg_primes = standard_dg_primes.copy()
        for i, rxn in enumerate(pathway.reactions):
            physiological_dg_primes[i] += (
                rxn.physiological_dg_correction() * R * default_T
            )

        optimized_dg_primes = standard_dg_primes.copy()
        _dg_adj = (pathway.S.T @ self.ln_conc).values.squeeze()
        optimized_dg_primes += _dg_adj * R * default_T

        # add the calculated error values (due to the ΔG'0 uncertainty)
        if y is not None and pathway.dg_sigma is not None:
            dg_adjustment = pathway.dg_sigma @ y.squeeze()
        else:
            dg_adjustment = np.zeros(standard_dg_primes.shape)

        # adjust ΔGs to flux directions
        standard_dg_primes = pathway.I_dir @ standard_dg_primes
        physiological_dg_primes = pathway.I_dir @ physiological_dg_primes
        optimized_dg_primes = pathway.I_dir @ optimized_dg_primes
        dg_adjustment = pathway.I_dir @ dg_adjustment

        # all dG values are in units of RT, so we convert them to kJ/mol
        reaction_data = zip(
            pathway.reaction_ids,
            pathway.reaction_formulas,
            pathway.fluxes,
            standard_dg_primes,
            standard_dg_primes + dg_adjustment,
            physiological_dg_primes + dg_adjustment,
            optimized_dg_primes + dg_adjustment,
        )
        self._reaction_df = pd.DataFrame(
            data=list(reaction_data),
            columns=[
                "reaction_id",
                "reaction_formula",
                "flux",
                "original_standard_dg_prime",
                "standard_dg_prime",
                "physiological_dg_prime",
                "optimized_dg_prime",
            ],
        )

        lbs, ubs = pathway.bounds
        compound_data = zip(
            pathway.compound_ids,
            np.exp(self.ln_conc).flatten() * standard_concentration,
            lbs,
            ubs,
        )
        self._compound_df = pd.DataFrame(
            data=list(compound_data),
            columns=[
                "compound_id",
                "concentration",
                "lower_bound",
                "upper_bound",
            ],
        )

    @property
    def reaction_df(self) -> pd.DataFrame:
        """Get a DataFrame with all the reaction data.

        The columns are:
            reaction_id
            reaction_formula
            flux
            original_standard_dg_prime
            standard_dg_prime
            physiological_dg_prime
            optimized_dg_prime

        """
        return self._reaction_df

    @property
    def compound_df(self) -> pd.DataFrame:
        """Get a DataFrame with all the compound data.

        The columns are:
            compound_id
            concentration
            lower_bound
            upper_bound

        """
        return self._compound_df

    @property
    def reaction_ids(self) -> Iterable[str]:
        return self._reaction_df.reaction_id.__iter__()

    @property
    def compound_ids(self) -> Iterable[str]:
        return self._compound_df.compound_id.__iter__()

    def to_sbtab(self) -> SBtab.SBtabDocument:
        """Generate a report (in SBtab format)."""
        sbtabdoc = SBtab.SBtabDocument("report")

        # add a table with the optimized metabolite concentrations
        met_data = []
        for row in self.compound_df.itertuples():
            met_data.append(
                (
                    "concentration",
                    row.compound_id,
                    f"{row.concentration.m_as('M'):.3e}",
                )
            )
        met_df = pd.DataFrame(
            columns=["!QuantityType", "!Compound", "!Value"], data=met_data
        )
        met_sbtab = SBtab.SBtabTable.from_data_frame(
            met_df,
            table_id="Predicted concentrations",
            table_type="Quantity",
            unit="M",
        )
        sbtabdoc.add_sbtab(met_sbtab)

        # add a table with the optimized reaction Gibbs energies
        rxn_data = []
        for row in self.reaction_df.itertuples():
            rxn_data.append(
                (
                    "reaction gibbs energy",
                    row.reaction_id,
                    f"{row.optimized_dg_prime.m_as('kJ/mol'):.3e}",
                )
            )
        rxn_df = pd.DataFrame(
            columns=["!QuantityType", "!Reaction", "!Value"], data=rxn_data
        )
        rxn_sbtab = SBtab.SBtabTable.from_data_frame(
            rxn_df,
            table_id="Predicted Gibbs energies",
            table_type="Quantity",
            unit="kJ/mol",
        )
        sbtabdoc.add_sbtab(rxn_sbtab)

        return sbtabdoc



"""thermo_models contains tools for running MDF and displaying results."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import logging
from typing import Callable, Dict, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd
from cvxpy import Maximize, Minimize, Problem, Variable, norm2
from cvxpy.constraints.constraint import Constraint
from cvxpy.problems.objective import Objective
from equilibrator_api import Q_, ComponentContribution, R, Reaction, default_T
from equilibrator_cache import Compound
from scipy import stats

from . import Bounds
from .mdf_solution import PathwayMdfSolution
from .mdmc_solution import PathwayMdmcSolution
from .pathway import Pathway


class ThermodynamicModel(Pathway):
    """Container for doing pathway-level thermodynamic analysis."""

    def __init__(
        self,
        reactions: List[Reaction],
        fluxes: Q_,
        comp_contrib: Optional[ComponentContribution] = None,
        standard_dg_primes: Optional[Q_] = None,
        dg_sigma: Optional[Q_] = None,
        bounds: Optional[Bounds] = None,
        config_dict: Optional[Dict[str, str]] = None,
        compound_id_mapping: Callable[[Compound], str] = None,
    ) -> None:
        """Initialize a Pathway object.

        Parameters
        ----------
        reactions : List[Reaction]
            a list of Reaction objects
        fluxes : Quantity
            relative fluxes in same order as
        comp_contrib : ComponentContribution
            a ComponentContribution object
        standard_dg_primes : Quantity, optional
            reaction energies (in kJ/mol)
        dg_sigma : Quantity, optional
            square root of the uncertainty covariance matrix (in kJ/mol)
        bounds : Bounds, optional
            bounds on metabolite concentrations (by default uses the
            "data/cofactors.csv" file in `equilibrator-api`)
        config_dict : dict, optional
            configuration parameters for Pathway analysis
        compound_id_mapping : callable, optional
            a function mapping compounds to their names in the model
        """
        super(ThermodynamicModel, self).__init__(
            reactions=reactions,
            fluxes=fluxes,
            comp_contrib=comp_contrib,
            standard_dg_primes=standard_dg_primes,
            dg_sigma=dg_sigma,
            bounds=bounds,
            config_dict=config_dict,
            compound_id_mapping=compound_id_mapping,
        )

    def _thermo_constraints(
        self,
        ln_conc: Variable,
        B: Variable,
    ) -> Tuple[Variable, List[Constraint], List[Constraint]]:
        """Create primal LP problem for Min-max Thermodynamic Driving Force.

        Returns
        -------
        the linear problem object, and the three types of variables as arrays.
        """

        _rt = (R * default_T).m_as("kJ/mol")
        dg_prime = (
            self.standard_dg_primes.m_as("kJ/mol")
            + _rt * self.S.T.values @ ln_conc
        )

        if self.dg_sigma is not None and self.dg_confidence > 0.0:
            # define the ΔG'0 covariance eigenvariables
            Nq = self.dg_sigma.shape[1]
            y = Variable(shape=Nq, name="covariance eigenvalues")
            # y_constraints = [-self.stdev_factor <= y, y <= self.stdev_factor]
            y_constraints = [
                norm2(y) <= stats.chi2.ppf(self.dg_confidence, Nq) ** (0.5)
            ]
            dg_prime += self.dg_sigma.m_as("kJ/mol") @ y
        else:
            # when the covariance is not provided, or the confidence interval
            # is 0.0 (which means we only allow the exact mean values as possible
            # assignments to ΔG'0), we still add need to a placeholder variable
            # named 'y' (otherwise the solver raises an exception).
            y = None
            y_constraints = []

        dg_constraints = []
        for i, direction in enumerate(np.diag(self.I_dir)):
            if direction != 0:
                dg_constraints += [direction * dg_prime[i] <= -B]
            else:
                dg_constraints += [direction * dg_prime[i] <= 1]  # placeholder

        return y, y_constraints, dg_constraints

    def _conc_constraints(
        self,
        ln_conc: Variable,
        constant_only: bool = False,
    ) -> Tuple[List[Constraint], List[Constraint]]:
        """Add lower and upper bounds for the log concentrations.

        Arguments
        ---------
        ln_conc : Variable
            the log-concentration variable, size should be (self.Nc, )
        constant_only : bool
            if True, only add bounds to the "constant" compounds, i.e. ones
            with 0 standard deviation
        """
        c_lbs = []
        c_ubs = []
        for j in range(self.Nc):
            if constant_only and self.ln_conc_sigma[j] > self.MINIMAL_STDEV:
                continue
            ci = self.ln_conc_sigma[j] * stats.chi2.ppf(
                self.ln_conc_confidence, 1
            ) ** (0.5)
            lb = self.ln_conc_mu[j] - ci
            ub = self.ln_conc_mu[j] + ci
            c_lbs += [ln_conc[j] >= lb]
            c_ubs += [ln_conc[j] <= ub]
        return c_lbs, c_ubs

    def mdf_analysis(self) -> PathwayMdfSolution:
        """Find the MDF (Max-min Driving Force).

        Returns
        -------
        a PathwayMDFData object with the results of MDF analysis.
        """
        # ln-concentration variables (where the units are in M before taking
        # the log)
        ln_conc = Variable(shape=self.Nc, name="log concentrations")
        c_lbs, c_ubs = self._conc_constraints(ln_conc)

        # the margin variable representing the MDF in units of kJ/mol
        B = Variable(shape=1, name="minimum driving force")
        y, y_constraints, dg_constraints = self._thermo_constraints(ln_conc, B)

        objective = Maximize(B)

        prob = Problem(
            objective, y_constraints + dg_constraints + c_lbs + c_ubs
        )
        prob.solve()
        if prob.status != "optimal":
            logging.warning("LP status %s", prob.status)
            raise Exception("Cannot solve MDF optimization problem")

        reaction_prices = np.array(
            [float(c.dual_value) for c in dg_constraints]
        ).round(5)
        compound_prices = np.array(
            [float(u.dual_value - l.dual_value) for l, u in zip(c_lbs, c_ubs)]
        ).round(5)

        return PathwayMdfSolution(
            self,
            score=prob.value,
            ln_conc=ln_conc.value,
            y=y.value if y is not None else None,
            reaction_prices=reaction_prices,
            compound_prices=compound_prices,
        )

    def get_zscores(self, ln_conc: Iterable) -> Iterable:
        return map(
            lambda x: (x[0] - x[1]) / x[2] if x[2] > self.MINIMAL_STDEV else 0,
            zip(ln_conc, self.ln_conc_mu, self.ln_conc_sigma),
        )

    def _make_zscore_objective(self, ln_conc: Variable) -> Objective:
        """Set the Z-score as the new objective."""
        zscores = [z ** 2 for z in self.get_zscores(ln_conc)]
        return Minimize(sum(zscores))

    def mdmc_analysis(
        self,
        min_lb: float = 0.0,
        max_lb: Optional[float] = 10.0,
        n_steps: int = 100,
    ) -> PathwayMdmcSolution:
        """Find the MDMC (Maximum Driving-force and Metabolic Consistency.

        :return: a PathwayMdmcSolution object with the results of MDMC analysis.
        """
        # ln-concentration variables (where the units are in M before taking
        # the log)
        ln_conc = Variable(shape=self.Nc, name="log concentrations")
        c_lbs, c_ubs = self._conc_constraints(ln_conc, constant_only=True)

        # the margin variable representing the MDF in units of kJ/mol
        B = Variable(shape=1, name="minimum driving force")
        y, y_constraints, dg_constraints = self._thermo_constraints(ln_conc, B)

        objective = self._make_zscore_objective(ln_conc)

        # scan through a range of DF lower bounds to find all possible Pareto
        # optimal solutions to the bi-optimization problem (MDF and Z-score)
        data = []
        for lb in np.linspace(min_lb, max_lb, n_steps):
            prob = Problem(
                objective,
                y_constraints + dg_constraints + c_lbs + c_ubs + [B >= lb],
            )
            prob.solve(solver="OSQP")
            if prob.status != "optimal":
                logging.warning("LP status %s", prob.status)
                raise Exception("Cannot solve MDF optimization problem")

            data.append((lb, "primal", "obj", "mdmc", 0, prob.value))
            data += [
                (lb, "primal", "var", "log_conc", i, ln_conc.value[i])
                for i in range(self.Nc)
            ]
            if self.dg_sigma is not None:
                Nq = self.dg_sigma.shape[1]
                data += [
                    (
                        lb,
                        "primal",
                        "var",
                        "covariance_eigenvalue",
                        i,
                        y.value[i] if y is not None else None,
                    )
                    for i in range(Nq)
                ]
            data += [
                (
                    lb,
                    "shadow_price",
                    "cnstr",
                    "driving_force",
                    j,
                    float(dg_constraints[j].dual_value),
                )
                for j in range(self.Nr)
            ]
            zscores = self.get_zscores(ln_conc.value)
            for j, zscore in enumerate(zscores):
                data.append((lb, "zscore", "var", "log_conc", j, zscore))

        solution_df = pd.DataFrame(
            data=data,
            columns=[
                "df_lb",
                "value_type",
                "var_type",
                "var_name",
                "index",
                "value",
            ],
        )
        return PathwayMdmcSolution(self, solution_df)



"""thermo_models contains tools for running MDF and displaying results."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Tuple

import numpy as np
import pandas as pd
from equilibrator_api import Q_
from matplotlib import pyplot as plt
from sbtab import SBtab

from .analysis_solution import PathwayAnalysisSolution
from .cost_function import EnzymeCostFunction
from .util import PlotCorrelation


class PathwayEcmSolution(PathwayAnalysisSolution):
    """Handle MDF results.

    PathwayMDFData is a container class for MDF results, with plotting
    capabilities.
    """

    def __init__(
        self, ecm_model: "EnzymeCostModel", score: float, ln_conc: np.ndarray
    ) -> None:
        """

        Parameters
        ----------
        ecm_model : EnzymeCostModel
        ln_conc : ndarray
            log concentrations at ECM optimum
        """
        super(PathwayEcmSolution, self).__init__(
            ecm_model._thermo_model, score, ln_conc
        )
        self.ecf = ecm_model.ecf

        if ecm_model._val_df_dict is not None:
            self.meas_met_conc = ecm_model._get_measured_metabolite_conc()
            self.meas_enz_conc = ecm_model._get_measured_enzyme_conc()
        else:
            self.meas_met_conc = None
            self.meas_enz_conc = None

        self.enzyme_df = pd.DataFrame(
            data=self.costs.tolist(),
            columns=["capacity", "thermodynamic", "saturation", "allosteric"],
        )

        self.enzyme_df["concentration"] = [
            np.exp(x.value) * Q_("M") for x in self.ecf.ECF(self.ln_conc)
        ]

        self.enzyme_df.insert(
            0, "reaction_id", pd.Series(list(self.reaction_ids))
        )

    @property
    def driving_forces(self) -> np.ndarray:
        """Calculate the driving forces of all reactions."""
        return self.ecf._driving_forces(self.ln_conc)

    @property
    def volumes(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get the volume occupied by each enzyme and metabolite.

        Returns
        -------
        enz_vols : array
        met_vols : array
        """
        return self.ecf.get_volumes(self.ln_conc)

    @property
    def costs(self) -> np.ndarray:
        """

        Returns
        -------
        costs : array (2D)
            A matrix containing the enzyme costs separated to the 4 ECF
            factors (as columns).
            The first column is the ECF1 predicted concentrations in [M].
            The other columns are unitless (added cost, always > 1)
        """
        return self.ecf.get_enzyme_cost_partitions(self.ln_conc)

    def _get_volume_data_for_plotting(self):
        enz_vols, met_vols = self.volumes

        enz_data = sorted(zip(enz_vols.flat, self.reaction_ids), reverse=True)
        enz_vols, enz_labels = zip(*enz_data)
        enz_colors = [(0.5, 0.8, 0.3)] * len(enz_vols)

        met_data = zip(met_vols.flat, self.compound_ids)
        # remove H2O from the list and sort by descending volume
        met_data = sorted(filter(lambda x: x[1] != "h2o", met_data))
        met_vols, met_labels = zip(*met_data)
        met_colors = [(0.3, 0.5, 0.8)] * len(met_vols)

        return (
            enz_vols + met_vols,
            enz_labels + met_labels,
            enz_colors + met_colors,
        )

    def plot_volumes(self, ax: plt.axes) -> None:
        width = 0.8
        vols, labels, colors = self._get_volume_data_for_plotting()

        ax.bar(np.arange(len(vols)), vols, width, color=colors)
        ax.set_xticklabels(labels, size="medium", rotation=90)
        ax.set_ylabel("total weight [g/L]")

    def plot_volumes_pie(self, ax: plt.axes) -> None:
        vols, labels, colors = self._get_volume_data_for_plotting()
        ax.pie(vols, labels=labels, colors=colors)
        ax.set_title(f"total weight = {sum(vols):.3g} [g/L]")

    def plot_thermodynamic_profile(self, ax: plt.axes) -> None:
        """
        Plot a cumulative line plot of the dG' values given the solution
        for the metabolite levels. This was originally designed for showing
        MDF results, but is also a useful tool for ECM.
        """
        dgs = [0] + list((-self.driving_forces).flat)
        cumulative_dgs = np.cumsum(dgs)

        xticks = np.arange(0, len(cumulative_dgs)) - 0.5
        xticklabels = [""] + list(self.reaction_ids)
        ax.plot(cumulative_dgs)
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, rotation=45, ha="right")
        ax.set_xlim(0, len(cumulative_dgs) - 1)
        ax.set_xlabel("")
        ax.set_ylabel(r"Cumulative $\Delta_r G'$ (kJ/mol)", family="sans-serif")

    def plot_enzyme_demand_breakdown(
        self, ax: plt.Axes, top_level: int = 3, plot_measured: bool = True
    ) -> None:
        """
        A bar plot in log-scale showing the partitioning of cost between
        the levels of kinetic costs:
        1 - capacity
        2 - thermodynamics
        3 - saturation
        4 - allosteric
        """
        assert top_level in range(1, 5)

        # give all reactions with zero cost a base value, which we will
        # also set as the bottom ylim, which will simulate a "minus infinity"
        # when we plot it in log-scale
        costs = self.costs
        base = min(filter(None, costs[:, 0])) / 2.0
        idx_zero = costs[:, 0] == 0
        costs[idx_zero, 0] = base
        costs[idx_zero, 1:] = 1.0

        bottoms = np.hstack(
            [np.ones((costs.shape[0], 1)) * base, np.cumprod(costs, 1)]
        )
        steps = np.diff(bottoms)

        labels = EnzymeCostFunction.ECF_LEVEL_NAMES[0:top_level]

        ind = range(costs.shape[0])  # the x locations for the groups
        width = 0.8
        ax.set_yscale("log")

        if plot_measured and self.meas_enz_conc is not None:
            ax.plot(
                ind,
                self.meas_enz_conc.m_as("M"),
                label="measured",
                color="gold",
                marker="d",
                markersize=7,
                linewidth=0,
                markeredgewidth=0.3,
                markeredgecolor=(0.3, 0.3, 0.3),
            )
        colors = ["tab:blue", "tab:orange", "tab:brown"]
        for i, label in enumerate(labels):
            ax.bar(
                ind,
                steps[:, i].flat,
                width,
                bottom=bottoms[:, i].flat,
                color=colors[i],
                label=label,
            )

        ax.set_xticks(ind)
        ax.set_xticklabels(self.reaction_ids, size="medium", rotation=90)
        ax.legend(loc="best", framealpha=0.2)
        ax.set_ylabel("enzyme demand [M]")
        ax.set_ylim(bottom=base)

    def validate_metabolite_conc(
        self, ax: plt.Axes, scale: str = "log"
    ) -> None:
        pred_met_conc = self.compound_df.concentration.apply(
            lambda x: x.m_as("M")
        ).values

        # remove NaNs and zeros
        mask = np.nan_to_num(self.meas_met_conc) > 0
        mask &= np.nan_to_num(pred_met_conc) > 0

        # remove compounds with fixed concentrations
        mask &= self.ecf.ln_conc_sigma > self.ecf.MINIMAL_STDEV

        PlotCorrelation(
            ax,
            self.meas_met_conc.m_as("M"),
            pred_met_conc,
            labels=self.compound_ids,
            mask=mask,
            scale=scale,
        )
        ax.set_xlabel("measured [M]")
        ax.set_ylabel("predicted [M]")

    def validate_enzyme_conc(self, ax: plt.Axes, scale: str = "log") -> None:
        pred_enz_conc = self.enzyme_df.concentration.apply(
            lambda x: x.m_as("M")
        ).values

        PlotCorrelation(
            ax,
            self.meas_enz_conc.m_as("M"),
            pred_enz_conc,
            labels=self.enzyme_df.reaction_id,
            scale=scale,
        )

        ax.set_xlabel("measured [M]")
        ax.set_ylabel("predicted [M]")

    def to_sbtab(self) -> SBtab.SBtabDocument:
        sbtabdoc = super(PathwayEcmSolution, self).to_sbtab()

        # add another table containing the optimized enzyme concentrations
        enz_df = self.enzyme_df[["reaction_id", "concentration"]].copy()
        enz_df["concentration"] = enz_df.concentration.apply(
            lambda x: f"{x.m_as('M'):.3e}"
        )
        enz_df.insert(0, "!QuantityType", "concentration of enzyme")
        enz_df = enz_df.rename(
            columns={"reaction_id": "!Reaction", "concentration": "!Value"}
        )

        enz_sbtab = SBtab.SBtabTable.from_data_frame(
            enz_df,
            table_id="Predicted enzyme levels",
            table_type="Quantity",
            unit="M",
        )
        sbtabdoc.add_sbtab(enz_sbtab)
        return sbtabdoc
